<?php
/**
 *  Flash Accel Management Console (FMC)
 *  Copyright (C) 2013, NetApp, Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Javascript
 *
 *    Farbtastic - Copyright (c) 2010 Matt Farina
 *
 *    jQuery - Copyright (c) 2010 John Resig
 * 
 *    jQuery BBQ - Copyright (c) 2010 "Cowboy" Ben Alman
 *  
 *    jQuery Cookie - Copyright (c) 2006 Klaus Hartl
 *  
 *    jQuery Form - Copyright (c) 2010 Mike Alsup
 *  
 *    jQuery Once - Copyright (c) 2009 Konstantin K\ufffdfer
 *
 *    jQuery UI - Copyright (c) 2010 by the original authors
 *      (http://jqueryui.com/about)
 *
 *    Sizzle.js - Copyright (c) 2010 The Dojo Foundation (http://sizzlejs.com/)
 *
 *  PHP
 *
 *    ArchiveTar - Copyright (c) 1997 - 2008 Vincent Blavet
 */

/**
 *  Open log file (truncating previous log file if it exists)
 */
function mercury_management_rediscover_open_logfile($type = '') {
  if ($type == '') {
    $type = 'discovery';
  }
  $file = mercury_management_get_log_dir() . "/$type.txt";
  $fh = fopen($file, "w");
  chmod($file, 0777);
  return $fh;
}


/**
 *  Write string to log file
 */
function mercury_management_rediscover_log($fh, $str) {
  if ($fh) {
     fwrite($fh, $str);
     fflush($fh);
  }
}


/**
 *  Close log file
 */
function mercury_management_rediscover_close_logfile($fh) {
   if ($fh) {
      fclose($fh);
   }
}


/**
 *  Check if user has cancelled discovery.
 */
function mercury_management_rediscover_check_cancel() {
  $vcInfo = mercury_management_vcenter_details();
  $rediscover_status = $vcInfo->field_rediscover_status['und'][0]['value'];
  if ($rediscover_status == 'rediscover_stop') {
    mercury_utilities_fmc_log_message('Discovery', 'Discovery terminated');
    $vcInfo->field_rediscover_status['und'][0]['value'] = 'rediscover_terminated';
    node_save($vcInfo);
    return false;
  }
  return true;
}


/**
 *  Delete a list of VM nodes with progress bar updates
 */
function mercury_management_delete_vms_and_update_ticks($log_fh, $vm_nids)
{
   mercury_management_rediscover_log($log_fh, "Removing obsolete vm nodes from the database.\n");
   foreach ($vm_nids as $nid) {
      mercury_management_rediscover_log($log_fh, "Deleting obsolete vm node id $nid and dependent nodes.\n");
      // Null is "don't log"; too verbose the way we're doing it
      mercury_management_delete_vms(null, array($nid));
      mercury_utilities_discovery_progress_register_ticks(1);
      if (false == mercury_management_rediscover_check_cancel()) {
         return false;
      }
   }
   return true;
}


/**
 *  Delete a list of host nodes with progress bar updates
 */
function mercury_management_delete_hosts_and_update_ticks($log_fh, $host_nids)
{
   mercury_management_rediscover_log($log_fh, "Removing obsolete host nodes from the database.\n");
   foreach ($host_nids as $nid) {
      mercury_management_rediscover_log($log_fh, "Deleting obsolete host node id $nid and dependent nodes.\n");
      // Null is "don't log"; too verbose the way we're doing it
      mercury_management_delete_hosts(null, array($nid));
      mercury_utilities_discovery_progress_register_ticks(1);
      if (false == mercury_management_rediscover_check_cancel()) {
         return false;
      }
   }
   return true;
}


/**
 *  UoW starting callback (for updating discovery pop-up textbox)
 */
function mercury_management_rediscover_uow_starting_cb(&$uow)
{
   $log_fh = $uow['log_fh'];
   if (isset($uow['vm'])) {
      mercury_management_rediscover_vm_log($uow, "starting discovery\n");
   } elseif (isset($uow['host_name'])) {
      mercury_management_rediscover_host_log($uow, "starting discovery\n");
   }
}


/**
 *  UoW finished callback (for updating progress bar and discovery pop-up textbox)
 */
function mercury_management_rediscover_uow_finished_cb(&$uow)
{
   $log_fh = $uow['log_fh'];
   if (isset($uow['vm'])) {
      mercury_management_rediscover_vm_log($uow, "completed discovery\n");
   } elseif (isset($uow['host_name'])) {
      mercury_management_rediscover_host_log($uow, "completed discovery\n");
   }
   mercury_utilities_discovery_progress_register_ticks(1);   
}


/**
 *  Rediscover entire system.
 */
function mercury_management_rediscover_body($log_fh, $vCenter, $vc_nid, $vc_hostname, $vc_username, $vc_password) {
  
  # Using 10 workers for now; semi-arbitrary, but 20 doesn't do much 
  # better than 10 (whereas 10 does a lot better than 1).
  $num_worker_tasks = 10;
  $discovery_error = 0;
  
  mercury_management_rediscover_log($log_fh, "Starting discovery\n");

  # Load the vCenter node; bail if there isn't one.
  db_query("RESET QUERY CACHE");
  $vcenters = db_query("SELECT SQL_NO_CACHE nid FROM {node} WHERE type = :tp AND status = :st", array(':tp' => 'vcenter', ':st' => 1));
  if (empty($vcenters)) {
    mercury_management_rediscover_log($log_fh, "Terminating discovery: no vCenter nodes!\n");
    mercury_utilities_fmc_log_message('Discovery', t('No vCenter nodes in database!'),  WATCHDOG_ERROR);
    return "";
  }

  # Create worker(s) to do the heavy lifting
  $wrkr_array = mercury_management_create_worker_array(
    $num_worker_tasks, 
    $vc_hostname,
    $vc_username,
    $vc_password
  );
  if ($wrkr_array === false) {
    mercury_management_rediscover_log($log_fh, "Terminating discovery: could not create worker array\n");
    mercury_utilities_fmc_log_message('Discovery', t('Could not connect to vCenter server.  Possible causes include invalid credentials, network problem, or server is down.'),  WATCHDOG_ERROR);
    $discovery_error++;
    goto done;
  }

  # Verify that this vCenter is running a supported software version
  $vcv = mercury_management_wa_get_vcversion($wrkr_array);
  if (FALSE == mercury_utilities_vcenter_version_is_supported($vcv)) {
    mercury_management_rediscover_log($log_fh, "Terminating discovery: vCenter $vcv not supported\n");
    mercury_utilities_fmc_log_message('Discovery', t("vCenter version $vcv is not supported by this version of FMC."),  WATCHDOG_ERROR);
    $discovery_error++;
    goto done;
  }

  # Set vc to the last vcenter node, in case there is more than one (preserve 
  # legacy behavior).
  foreach ($vcenters as $vc) {}

  # Note: this looks more appealing but doesn't work.
  # $vc = end($vcenters); 

  # Since we can manage only 1 vCenter now, there will be only 1 assignment to 
  # datacenter_names.  If FMC chooses to maintain a list of vCenters, then the 
  # following variable should be an array of arrays.  
  $datacenter_names = mercury_management_wa_get_dclist($wrkr_array);

  # Ensure datacenter nodes are synchronized with vSphere
  mercury_management_synchronize_datacenter_nodes($datacenter_names, $vc->nid, $log_fh);

  # Extract the subset of the datacenters that have checkmarks.
  $selected_datacenters = mercury_management_extract_selected_datacenters($datacenter_names);

  # Obtain the lists of hosts and VM nodes to be deleted
  $valid_hosts = array();
  $valid_vms = array();
  mercury_management_find_hosts_and_vms_in_datacenters($log_fh, $wrkr_array, $selected_datacenters, $valid_hosts, $valid_vms);
  $obsolete_host_nids = mercury_management_find_obsolete_hosts($log_fh, $valid_hosts);
  $obsolete_vm_nids = mercury_management_find_obsolete_vms($log_fh, $valid_vms);
  
  # Count how many host and VM nodes need to be deleted
  $hosts_and_vms_to_delete = count($obsolete_host_nids) + count($obsolete_vm_nids);
  
  # Obtain array of hosts and VMs to be discovered
  $vCenterArr = variable_get('rediscover_items');
  
  # Count how many hosts and VMs need to be discovered
  $hosts_and_vms_to_discover = 0;
  foreach ($vCenterArr as $item) {
    if (($item['type'] == 'host') || ($item['type'] == 'vm')) {
      $hosts_and_vms_to_discover++;
    }
  }

  # This bit is just to make the progress bar move a bit while the annotations 
  # are read, since that takes a while.  Numbers were chosen empirically.
  $ndc = count($selected_datacenters);
  if ($ndc) {
     $ticks_for_annotation = intval(($hosts_and_vms_to_discover)/10);
     $ticks_for_annotation_per_dc = intval($ticks_for_annotation/$ndc);
  } else {
     $ticks_for_annotation = 0;
     $ticks_for_annotation_per_dc = 0;  
  }
  # Initialize the information needed for the progress bar to work
  mercury_management_rediscover_log($log_fh, "hosts and vms to delete   : $hosts_and_vms_to_delete\n");
  mercury_management_rediscover_log($log_fh, "hosts and vms to discover : $hosts_and_vms_to_discover\n");
  $total_ticks = $hosts_and_vms_to_delete + $ticks_for_annotation + $hosts_and_vms_to_discover;
  mercury_utilities_discovery_progress_init_ticks($total_ticks);

  # Perform host and VM deletions
  if (false == mercury_management_delete_vms_and_update_ticks($log_fh, $obsolete_vm_nids)) {
    mercury_management_rediscover_log($log_fh, "vm deletion terminated prematurely!\n");
    $discovery_error++;
    goto done;
  }
  if (false == mercury_management_delete_hosts_and_update_ticks($log_fh, $obsolete_host_nids)) {
    mercury_management_rediscover_log($log_fh, "host deletion terminated prematurely!\n");
    $discovery_error++;
    goto done; 
  }
  
  # If nothing new to discover, we're done
  if (empty($vCenterArr)) {
    mercury_management_rediscover_log($log_fh, "Terminating discovery: no items to discover\n");
    goto done;
  }
  
  # Assemble lists of annotated hosts and VMs in selected datacenters.
  $annotatedVMs = array();
  $annotatedHosts = array();
  foreach ($selected_datacenters as $dataCenter) {
    mercury_utilities_discovery_progress_register_ticks($ticks_for_annotation_per_dc);
    $ticks_for_annotation -= $ticks_for_annotation_per_dc;
    $annotatedVMList = array();
    $getAnnotatedListResult =  mercury_management_list_annotated_components('vm', $dataCenter, $annotatedVMList);
    if ($getAnnotatedListResult === false) {
      mercury_management_rediscover_log($log_fh, "Annotations VM error ". $annotatedVMList . "\n");
      break;
    }

    $annotatedHostList = array();
    $getAnnotatedListResult = mercury_management_list_annotated_components('host', $dataCenter, $annotatedHostList);
    if ($getAnnotatedListResult === false) {
       mercury_management_rediscover_log($log_fh, "Annotations Host error ". $annotatedHostList . "\n");
      break;
    }

    foreach ($annotatedVMList as $annotatedVM) {
      mercury_management_rediscover_log($log_fh, "Annotated VM ". $annotatedVM . "\n");
      $annotatedVMs[$annotatedVM] = true;
    }

    foreach ($annotatedHostList as $annotatedHost) {
      mercury_management_rediscover_log($log_fh, "Annotated Host ". $annotatedHost . "\n");
      $annotatedHosts[$annotatedHost] = true;
    }
  }
  if ($ticks_for_annotation > 0) {
    mercury_utilities_discovery_progress_register_ticks($ticks_for_annotation);
  }
  
  mercury_management_rediscover_log($log_fh, "Resetting query cache.\n");
  db_query("RESET QUERY CACHE");

  # Assemble the host UoW queue
  mercury_management_rediscover_log($log_fh, "Assembling host UoW queue.\n");
  $host_uow_queue = array();
  foreach ($vCenterArr as $item) {
    if ($item['type'] == "host") {
      if (array_key_exists($item['name'], $annotatedHosts) === false) {
        $annotatedHosts[$item['name']] = false;
      }

      $host_name = trim($item['name']);
      $host_uow = mercury_management_assemble_host_uow(
        $log_fh,
        $vc_hostname, 
        $vc_username, 
        $vc_password, 
        $host_name, 
        $item['datacenter_id'], 
        $annotatedHosts[$item['name']],
        FALSE
      );
      $host_uow['starting_cbk'] = 'mercury_management_rediscover_uow_starting_cb';
      $host_uow['finished_cbk'] = 'mercury_management_rediscover_uow_finished_cb';
      $host_uow_queue[] = $host_uow;
    }
  }
  
  # Discover the hosts
  mercury_management_rediscover_log($log_fh, "Initiating host discovery.\n");
  $continue_cb = 'mercury_management_rediscover_check_cancel';
  if (mercury_management_process_uow_queue($host_uow_queue, $wrkr_array, $continue_cb)) {
    mercury_management_rediscover_log($log_fh, "Completed host discovery.\n");
  } else {
    mercury_management_rediscover_log($log_fh, "Host discovery terminated prematurely!\n");
    $discovery_error++;
    goto done;
  }
  
  # Extract the host nids (need for VM linkage currently) and look for
  # any discovery errors.
  $host_nids = array();
  foreach ($host_uow_queue as &$uow) {
    if (isset($uow['host_info'])) {
      $host_info = $uow['host_info'];
      $host_name = $uow['host_name'];
      $host_nids[$host_name] = $host_info->nid;
    }
    if (isset($uow['error'])) {
      $discovery_error++;
    }
  }
  
  # Free the host UoW queue
  unset($host_uow_queue);
  
  # Get the current set of VMs and the corresponding hosts
  mercury_management_rediscover_log($log_fh, "Obtaining list of hosts and VMs.\n");
  $dataArr = mercury_management_wa_get_vmlist($wrkr_array);
  $currentVC = array();
  if (!empty($dataArr)) {
    foreach ($dataArr as $item) {
      $host_split = explode('Host :', $item);
      if (count($host_split) > 1) {
        $currHost = trim($host_split[1]);
      }
      $vm_split = explode('VM :', $item);
      if (count($vm_split) > 1) {
        $vm = trim($vm_split[1]);
        $currentVC[$vm] = $currHost;
      }
    }
  }
  
  # Assemble the VM UoW queue
  $vm_uow_queue = array();
  foreach ($vCenterArr as $item) {
    if ($item['type'] == "vm") {
      if (array_key_exists($item['name'], $annotatedVMs) === false) {
        $annotatedVMs[$item['name']] = false;
      }

      $vmName = $item['name'];
      $vm_uow = mercury_management_assemble_vm_uow(
        $log_fh,
        $vc_hostname, 
        $vc_username, 
        $vc_password, 
        $vmName,
        $annotatedVMs[$item['name']], 
        $host_nids[$currentVC[$vmName]]
      );
      $vm_uow['starting_cbk'] = 'mercury_management_rediscover_uow_starting_cb';
      $vm_uow['finished_cbk'] = 'mercury_management_rediscover_uow_finished_cb';
      $vm_uow_queue[] = $vm_uow;
    }
  }

  # Discover the VMs
  mercury_management_rediscover_log($log_fh, "Initiating VM discovery.\n");
  $continue_cb = 'mercury_management_rediscover_check_cancel';
  if (mercury_management_process_uow_queue($vm_uow_queue, $wrkr_array, $continue_cb)) {
    mercury_management_rediscover_log($log_fh, "Completed VM discovery.\n");
  } else {
    mercury_management_rediscover_log($log_fh, "VM discovery terminated prematurely!\n");
    $discovery_error++;
    goto done;
  }
  
  # Look for any discovery errors
  foreach ($vm_uow_queue as &$uow) {
    if (isset($uow['error'])) {
      $discovery_error++;
    }
  }

  mercury_utilities_fmc_log_message('Discovery', t('Finished discovering all entities.'));

done:
    
  $vcInfo = mercury_management_vcenter_details();
  if ($vcInfo->field_rediscover_status['und'][0]['value'] == 'rediscover_terminated') {
    mercury_management_rediscover_log($log_fh, "Discovery was cancelled before completion.\n");
  } 
  else {
    if ($discovery_error == 0) {
      $vcInfo->field_rediscover_status['und'][0]['value'] = '';
    } else {
      $vcInfo->field_rediscover_status['und'][0]['value'] = 'rediscover_failed';
    }
    mercury_management_rediscover_log($log_fh, "Completed discovery with $discovery_error errors.\n");
    node_save($vcInfo);
    # Clear the node cache and reload this node, just to be safe
    node_load($vcInfo->nid, NULL, TRUE);
  }    
  
  # Make sure to clean up the worker pool
  mercury_management_destroy_worker_array($wrkr_array);
  
  return "";
}


/**
 *  Rediscover entire system.
 */
function mercury_management_rediscover($vCenter, $vc_nid, $vc_hostname, $vc_username, $vc_password) {

  # Open log file
  $log_fh = mercury_management_rediscover_open_logfile('discovery');

  # Execute discovery in a try block
  try {
     mercury_management_rediscover_body($log_fh, $vCenter, $vc_nid, $vc_hostname, $vc_username, $vc_password);
  }
  catch (Exception $e) {
     $msg = $e->getMessage();
     mercury_management_rediscover_log($log_fh, "Discovery terminated by exception: $msg\n");
     mercury_utilities_fmc_log_message('Discovery', t("Discovery terminated by exception: $msg"),  WATCHDOG_ERROR);
     try {
       $vcenters = db_query("SELECT SQL_NO_CACHE nid FROM {node} WHERE type = :tp AND status = :st", array(':tp' => 'vcenter', ':st' => 1));
     }
     catch (Exception $e) {
       $msg = $e->getMessage();
       mercury_management_rediscover_log($log_fh, "db_query failed due to exception: $msg\n");
       $vcenters = array();
     }
     if (!empty($vcenters)) {
        $vcInfo = node_load(end($vcenters)->nid);
        $vcInfo->field_rediscover_status['und'][0]['value'] = 'rediscover_failed';
        node_save($vcInfo);
        node_load($vcInfo_nid, NULL, TRUE);
        mercury_management_rediscover_log($log_fh, "Set rediscover_failed\n");
     } else {
        mercury_management_rediscover_log($log_fh, "Couldn't set rediscover_failed!\n");
     }
  }

  # Close the log file
  mercury_management_rediscover_close_logfile($log_fh);
  
  return "";
}


/**
 *  Function to created & update a virtual machine
 */
function mercury_management_updatevm($vc_hostname, $vc_username, $vc_password, $vm, $host_nid) {

  # Initialize return value
  $rv = '';
  
  # Open log file
  $log_fh = mercury_management_rediscover_open_logfile('refresh');

  mercury_management_rediscover_log($log_fh, "Starting refresh of VM $vm (host nid $host_nid)\n");

  # Create worker to do the heavy lifting (primary advantage: persistent 
  # connection to vCenter server).  No need for more than 1 given how
  # parallelism is currently accomplished.
  $num_worker_tasks = 1;
  $wrkr_array = mercury_management_create_worker_array(
    $num_worker_tasks, 
    $vc_hostname,
    $vc_username,
    $vc_password
  );
  if ($wrkr_array === false) {
    mercury_management_rediscover_log($log_fh, "VM $vm refresh: Could not create worker array!\n");
    mercury_utilities_fmc_log_message('VM Refresh', t('Could not connect to vCenter server.  Possible causes include invalid credentials, network problem, or server is down.'),  WATCHDOG_ERROR);
    mercury_management_rediscover_close_logfile($log_fh);
    return 'error--';
  }

  # Assemble the VM UoW queue (just one UoW in it)
  $uow = mercury_management_assemble_vm_uow(
    $log_fh,
    $vc_hostname, 
    $vc_username, 
    $vc_password, 
    $vm, 
    false, # Always redo annotations on a refresh
    $host_nid
  );
  $vm_uow_queue = array();
  $vm_uow_queue[0] = &$uow;
  
  # Refresh the VM
  if (mercury_management_process_uow_queue($vm_uow_queue, $wrkr_array)) {
    if (isset($uow['error'])) {
      $err = $uow['error'];
      mercury_management_rediscover_log($log_fh, "VM $vm refresh error: $err.\n");
      $rv .= 'error--';
    } else {
      mercury_management_rediscover_log($log_fh, "Completed VM $vm refresh.\n");
    }
  } else {
    mercury_management_rediscover_log($log_fh, "VM $vm refresh terminated prematurely!\n");
    $rv .= 'error--';
  }
    
  # Make sure to clean up the worker pool
  mercury_management_destroy_worker_array($wrkr_array);
  
  # Append VM nid to our result (this is the only result in the success case)
  if (isset($uow['vm_info'])) {
    $vm_info = $uow['vm_info'];
    $rv .= $vm_info->nid;
  }
  
  # Close the log file
  mercury_management_rediscover_close_logfile($log_fh);

  return $rv;
}


/**
 *  Function to create & update a host
 */
function mercury_management_updatehost($vc_hostname, $vc_username, $vc_password, $hostname, $datacenter_nid, $is_installing) {

  # Initialize return value
  $rv = '';
  
  # Open log file
  $log_fh = mercury_management_rediscover_open_logfile('refresh');

  mercury_management_rediscover_log($log_fh, "Starting refresh of host $hostname (d/c nid $datacenter_nid)\n");

  # Create worker to do the heavy lifting (primary advantage: persistent 
  # connection to vCenter server).  No need for more than 1 given how
  # parallelism is currently accomplished.
  $num_worker_tasks = 1;
  $wrkr_array = mercury_management_create_worker_array(
    $num_worker_tasks, 
    $vc_hostname,
    $vc_username,
    $vc_password
  );
  if ($wrkr_array === false) {
    mercury_management_rediscover_log($log_fh, "Host $hostname refresh: Could not create worker array!\n");
    mercury_utilities_fmc_log_message('Host Refresh', t('Could not connect to vCenter server.  Possible causes include invalid credentials, network problem, or server is down.'),  WATCHDOG_ERROR);
    mercury_management_rediscover_close_logfile($log_fh);
    return 'error--';
  }

  # Assemble the host UoW queue (just one UoW in it)
  $uow = mercury_management_assemble_host_uow(
    $log_fh,
    $vc_hostname, 
    $vc_username, 
    $vc_password, 
    $hostname, 
    $datacenter_nid,
    FALSE, // Force a check of the install status (ignore annotation)
    $is_installing
  );
  $host_uow_queue = array();
  $host_uow_queue[0] = &$uow;
  
  # Refresh the host
  if (mercury_management_process_uow_queue($host_uow_queue, $wrkr_array)) {
    if (isset($uow['error'])) {
      $err = $uow['error'];
      mercury_management_rediscover_log($log_fh, "Host $hostname refresh error: $err.\n");
      $rv .= 'error--';
    } else {
      mercury_management_rediscover_log($log_fh, "Completed host $hostname refresh.\n");
    }
  } else {
    mercury_management_rediscover_log($log_fh, "Host $hostname refresh terminated prematurely!\n");
    $rv .= 'error--';
  }
    
  # Make sure to clean up the worker pool
  mercury_management_destroy_worker_array($wrkr_array);
  
  # Append host nid to our result (this is the only result in the success case)
  if (isset($uow['host_info'])) {
    $host_info = $uow['host_info'];
    $rv .= $host_info->nid;
  }
  
  # Close the log file
  mercury_management_rediscover_close_logfile($log_fh);

  return $rv;
}
