<?php
/**
 *  Flash Accel Management Console (FMC)
 *  Copyright (C) 2013, NetApp, Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Javascript
 *
 *    Farbtastic - Copyright (c) 2010 Matt Farina
 *
 *    jQuery - Copyright (c) 2010 John Resig
 * 
 *    jQuery BBQ - Copyright (c) 2010 "Cowboy" Ben Alman
 *  
 *    jQuery Cookie - Copyright (c) 2006 Klaus Hartl
 *  
 *    jQuery Form - Copyright (c) 2010 Mike Alsup
 *  
 *    jQuery Once - Copyright (c) 2009 Konstantin K\ufffdfer
 *
 *    jQuery UI - Copyright (c) 2010 by the original authors
 *      (http://jqueryui.com/about)
 *
 *    Sizzle.js - Copyright (c) 2010 The Dojo Foundation (http://sizzlejs.com/)
 *
 *  PHP
 *
 *    ArchiveTar - Copyright (c) 1997 - 2008 Vincent Blavet
 */

define('FAMC_HOST_BUNDLE_STATUS_INSTALLED',      'Agent Installed');
define('FAMC_HOST_BUNDLE_STATUS_NOT_INSTALLED',  'Agent Not Installed');
define('FAMC_HOST_BUNDLE_STATUS_NOT_COMPATIBLE', 'Not Compatible');
define('FAMC_HOST_BUNDLE_STATUS_NOT_KNOWN',      'Unable to Communicate with Host');


/**
 *  Create or update the database entry for the given host.
 */
function mercury_management_create_or_update_host_node(&$uow, $host_info_with_key, $datacenter_nid)
{
  global $user;

  $action = "<undefined>";
  
  $nid = db_query("SELECT nid FROM {node} WHERE type = :tp AND title = :ttl", array(':tp' => 'host', ':ttl' => trim($host_info_with_key['Host Name'])))->fetchField();
  if (!empty($nid)) {
    $host_info = node_load($nid);
    $uow['host_info_orig'] = clone $host_info;
    $action = "updated";
  }
  else {
    $host_info = new stdClass();
    $host_info->type = 'host';
    $host_info->author = $user->name;
    $host_info->status = 1;
    $host_info->uid = $user->uid;
    $host_info->comment = 0;
    $host_info->language = 'und';
    $host_info->format = 2;
    $host_info->title = trim($host_info_with_key['Host Name']);
    $action = "created";
  }

  mercury_management_initialize_host_node($host_info);

  $host_info->field_cpu_model['und'][0]['value'] = $host_info_with_key['Cpu Model'];
  $host_info->field_cpu_usage['und'][0]['value'] = $host_info_with_key['Cpu Usage'];
  $host_info->field_boottime['und'][0]['value'] = $host_info_with_key['BootTime'];
  $host_info->field_host_status['und'][0]['value'] = $host_info_with_key['Host Status'];
  $host_info->field_maintenance_mode['und'][0]['value'] = $host_info_with_key['Maintenance mode'];
  $host_info->field_reboot_required['und'][0]['value'] = $host_info_with_key['Reboot Required'];
  $host_info->field_physical_memory['und'][0]['value'] = $host_info_with_key['Physical Memory'];
  $host_info->field_vmotion['und'][0]['value'] = $host_info_with_key['VMotion'];
  $host_info->field_drs['und'][0]['value'] = $host_info_with_key['DRS'];
  $host_info->field_ha['und'][0]['value'] = $host_info_with_key['HA'];
  $host_info->field_software_on_host['und'][0]['value'] = $host_info_with_key['Software on host'];
  $host_info->field_host_model['und'][0]['value'] = $host_info_with_key['Model'];
  $host_info->field_host_vendor['und'][0]['value'] = $host_info_with_key['Vendor'];
  $host_info->field_datacenter['und'][0]['target_id'] = $datacenter_nid;
  node_save($host_info);
  
  $uow['host_info'] = $host_info;
  if ($action == 'updated') {
    $uow['host_info_is_new'] = false;
  } else {
    $uow['host_info_is_new'] = true; 
  }

  $nid = $host_info->nid;
  mercury_management_rediscover_host_log($uow, "$action host nid $nid (dc_nid $datacenter_nid)\n");

  return $host_info;
}


/**
 * Update the bundle_status field on a phase A failure.  If the host is not in 
 * the database, create a node for it.  Otherwise use the original node values
 * except for bundle_status, so we can be sure not to leave the database in
 * an inconsistent state.
 */
function mercury_management_update_bundle_status_on_phase_a_failure(&$uow, $status)
{
   if (false == $uow['host_info_is_new']) {
      $host_info = $uow['host_info_orig'];
   } else {
      $host_info = $uow['host_info'];
   }
   $host_info->field_bundle_status['und'][0]['value'] = $status;
   node_save($host_info);
}


/**
 *  Add host_control.py 1 data to a host database node.
 */
function mercury_management_add_host_control_1_data_to_host_node($host_info_with_key, $host_info)
{
  $host_info->field_allocated_segments['und'][0]['value'] = $host_info_with_key['AllocatedSegments'];
  $host_info->field_segment_size['und'][0]['value'] = $host_info_with_key['SegmentSize'];
  $host_info->field_max_virtual_machines['und'][0]['value'] = $host_info_with_key['MaxVirtualMachines'];
  $host_info->field_max_physical_devices['und'][0]['value'] = $host_info_with_key['MaxPhysicalDevices'];
  $host_info->field_claimed_physical_devices['und'][0]['value'] = $host_info_with_key['ClaimedPhysicalDevices'];
  $host_info->field_cache_state['und'][0]['value'] = $host_info_with_key['CacheState'];
  $host_info->field_total_segments['und'][0]['value'] = $host_info_with_key['TotalSegments'];
  $host_info->field_allocated_virtual_machines['und'][0]['value'] = $host_info_with_key['AllocatedVirtualMachines'];
  $host_info->field_vib_version['und'][0]['value'] = $host_info_with_key['VibVersion'];
  $vmotion_supported = 0;
  if (array_key_exists ('Capabilities',  $host_info_with_key)) {
    $value = $host_info_with_key['Capabilities'];
    if (($value & 0x1) == 1) {
      $vmotion_supported = 1;
    }
  }
  $host_info->field_vmotion['und'][0]['value'] = $vmotion_supported;
  if (array_key_exists ('NAAIdentifier' , $host_info_with_key)) {
    $host_info->field_total_size['und'][0]['value'] = $host_info_with_key['TotalSize'];
    $host_info->field_naa_identifier['und'][0]['value'] = $host_info_with_key['NAAIdentifier'];

    if (empty($host_info_with_key['AllocatedSize'])) {
      $host_info->field_allocated_size['und'][0]['value'] = 0;
    } else {
      $host_info->field_allocated_size['und'][0]['value'] = $host_info_with_key['AllocatedSize'];
    }

    $host_info->field_device_state['und'][0]['value'] = $host_info_with_key['DeviceState'];
    $host_info->field_hbtl['und'][0]['value'] = $host_info_with_key['HBTL'];
  } else {
    $host_info->field_total_size['und'][0]['value'] = '';
    $host_info->field_naa_identifier['und'][0]['value'] = '';
    $host_info->field_allocated_size['und'][0]['value'] = '';
    $host_info->field_device_state['und'][0]['value'] = 0;
    $host_info->field_hbtl['und'][0]['value'] = '';
  }
}


/**
 *  Create or update the device database entries associated with a host.
 */
function mercury_management_update_host_device_db_commit(&$uow, $deviceArray, $host_nid)
{
  global $user;

  if (!is_numeric($host_nid)) {
    return false;
  }

  $device_nids = mercury_management_find_dependent_nodes(array($host_nid), 'device', 'field_related_host');
  $ndn = count($device_nids);
  
  # Optimize the case where we're deleting all host device nodes.
  if (empty($deviceArray)) {
    if ($ndn) {
      mercury_management_rediscover_host_log($uow, "deleting $ndn device nodes linked to host_nid $host_nid\n");
      node_delete_multiple($device_nids);
    } else {
      mercury_management_rediscover_host_log($uow, "couldn't find any device nodes linked to host_nid $host_nid\n");
    }
    return true;
  }

  # We have some host devices, some of which may already have nodes.
  
  # Load all existing nodes; we're going to delete the ones that don't get
  # updated.    
  if (isset($device_nids)) {
    $device_nodes = entity_load('node', $device_nids);
  } else {
    $device_nodes = array();
  }

  # Search for each of the passed-in devices in the existing device nodes.
  foreach ($deviceArray as $device) {
    $title = trim($device['Display Name']);
    $action = "creating";
    foreach ($device_nodes as $di_key => $device_info) {
      if ($device_info->title === $title) {
        $action = "updating";
        break;
      }
    } 
    if ($action == "creating") {
      $device_info = new stdClass();
      $device_info->type = 'device';
      $device_info->author = $user->name;
      $device_info->status = 1;
      $device_info->uid = $user->uid;
      $device_info->comment = 0;
      $device_info->language = 'und';
      $device_info->format = 2;
      $device_info->title = trim($device['Display Name']);
    } else {
      # Later we're going to delete everything left in device_nodes
      unset($device_nodes[$di_key]);
    }
    mercury_management_initialize_device_node($device_info);

    $device_info->field_related_host['und'][0]['target_id'] = $host_nid;
    $device_info->field_serial_no['und'][0]['value'] = $device['serial_no'];
    $device_info->field_display_name['und'][0]['value'] = $device['Display Name'];
    $device_info->field_size['und'][0]['value'] = $device['Size'];
    $device_info->field_device_type['und'][0]['value'] = $device['Device Type'];
    $device_info->field_multipath_plugin['und'][0]['value'] = $device['Multipath Plugin'];
    $device_info->field_devfs_path['und'][0]['value'] = $device['Devfs Path'];
    $device_info->field_vendor['und'][0]['value'] = $device['Vendor'];
    $device_info->field_model['und'][0]['value'] = $device['Model'];
    $device_info->field_revision['und'][0]['value'] = $device['Revision'];
    $device_info->field_scsi_level['und'][0]['value'] = $device['SCSI Level'];
    $device_info->field_is_pseudo['und'][0]['value'] = $device['Is Pseudo'];
    $device_info->field_status['und'][0]['value'] = $device['Status'];
    $device_info->field_is_rdm_capable['und'][0]['value'] = $device['Is RDM Capable'];
    $device_info->field_is_local['und'][0]['value'] = $device['Is Local'];
    $device_info->field_is_removable['und'][0]['value'] = $device['Is Removable'];
    $device_info->field_is_ssd['und'][0]['value'] = $device['Is SSD'];
    $device_info->field_is_offline['und'][0]['value'] = $device['Is Offline'];
    $device_info->field_is_perennially_reserved['und'][0]['value'] = $device['Is Perennially Reserved'];
    $device_info->field_thin_provisioning_status['und'][0]['value'] = $device['Thin Provisioning Status'];
    $device_info->field_attached_filters['und'][0]['value'] = $device['Attached Filters'];
    $device_info->field_vaai_status['und'][0]['value'] = $device['VAAI Status'];
    $device_info->field_other_uids['und'][0]['value'] = $device['Other UIDs'];
    node_save($device_info);
    
    $nid = $device_info->nid;
    $id = $device_info->title;
    mercury_management_rediscover_host_log($uow, "$action host device nid $nid (host_nid $host_nid) id $id\n");
  }

  # Any device still "set" in device_nodes needs to be deleted  
  $device_nids_to_delete = array();
  foreach ($device_nodes as $device_info) {
    $nid = $device_info->nid;
    $id = $device_info->title;
    mercury_management_rediscover_host_log($uow, "deleting host device nid $nid (host_nid $host_nid) id $id\n");
    $device_nids_to_delete[] = $device_info->nid;
  }
  if (!empty($device_nids_to_delete)) {
    node_delete_multiple($device_nids_to_delete);
  }
  
  return true;
}


/**
 *  Attempt to fix SFCBD on a host.
 */
function mercury_management_attempt_sfcbd_fix($vc_hostname, $vc_username, $vc_password, $host_name, $host_username, $host_password)
{
  /* Possible Scenario
   * 1. SFCBD is restarting. If thats the case, then wait for restart to complete before trying any recovery
   * 2. SFCBD is hung. This may mean it is either stopped are can't be stopped. Restart it then.
   *
   * Solution
   * 1. Poll and detect possible transition condition.
   * 2. If not transitioning, restart, and poll to connect.
   * 3. If you still can't connect then something seriously messed up.
   */
  $result = array();
  $currentState = mercury_management_handle_vcenter_discover_perl_script($vc_hostname, $vc_username, $vc_password, $host_name, "serviceGetSfcbdWatchdogStatus");
  if (stristr($currentState, "failed")) {
    mercury_utilities_fmc_log_message('Discovery', 'Cannot connect to host. Please check if it is powered on and try again.',  WATCHDOG_WARNING);
    return $result;
  }

  $previousState = $currentState;
  for($i = 0; $i < 10; $i++) {
    $currentState = mercury_management_handle_vcenter_discover_perl_script($vc_hostname, $vc_username, $vc_password, $host_name, "serviceGetSfcbdWatchdogStatus");
    if (stristr($currentState, "failed")) {
      return $result;
    }

    if ($currentState != $previousState) {
      /* State change in progress already */
      return $result;
    }
  }

  mercury_management_handle_vcenter_discover_perl_script($vc_hostname, $vc_username, $vc_password, $host_name, "restartSfcbdWatchdog");

  for ($i = 0; $i < 10; $i++) {
    $get_host_info2 = exec("/usr/lib/vmware-vcli/apps/host/host_control.py '" . trim($host_name) . "' '" . $host_username . "' '" . $host_password . "' 1", $result);
    $currentState = mercury_management_handle_vcenter_discover_perl_script($vc_hostname, $vc_username, $vc_password, $host_name, "serviceGetSfcbdWatchdogStatus");
    if (count($result) > 0) {
      /* Success */
      break;
    }
    $currentState = "failed";
  }

  if ($currentState != $previousState) {
    /* All the recover operations failed. Some serious problem with sfcbd */
    mercury_utilities_fmc_log_message('Host Refresh', 'Host is not in a stable state. Please check if services are configured correctly. A reboot may be required to restore state.', WATCHDOG_ERROR);
    return array();
  }
  
  return $result;
}


/**
 *  Decide whether host discovery should be aborted due to unacknowledged alarms.
 */
function mercury_management_abort_updatehost_due_to_pending_alarms(&$uow, $host_info_with_key)
{
  $host_info = $uow['host_info'];
  $pending_alarms = $host_info_with_key['Pending Alarms'];
  if ($pending_alarms == "Yes") {
    mercury_utilities_fmc_log_message('Host Refresh', t('Cannot connect to host @host. Please acknowledge pending alarms.', array('@host' => $host_info->title)), WATCHDOG_WARNING);
    mercury_management_update_bundle_status_on_phase_a_failure($uow, FAMC_HOST_BUNDLE_STATUS_NOT_KNOWN);
    return true;
  }
  else {
    return false;
  }
}


/**
 *  Decide whether host discovery should be aborted due to host agent software status.
 */
function mercury_management_abort_updatehost_due_to_bundle_status(&$uow, $bundle_status)
{
  $host_info = $uow['host_info'];
  if ($bundle_status == "Error") {
    mercury_utilities_fmc_log_message('Host Refresh', t('Username/Password incorrect for host @host.', array('@host' => $host_info->title)), WATCHDOG_ERROR);
  } else if ($bundle_status == "CannotConnect") {
    mercury_utilities_fmc_log_message('Host Refresh', t('Cannot connect to host @host. Host may be powered off.', array('@host' => $host_info->title)), WATCHDOG_WARNING);
  }
  if ($bundle_status == "Error" || $bundle_status == "CannotConnect") {
    mercury_management_update_bundle_status_on_phase_a_failure($uow, FAMC_HOST_BUNDLE_STATUS_NOT_KNOWN);
    return true;
  } else {
    return false;
  }
}


/**
 *  Log a string to the log file associated with this UoW, tagged with host name.
 */
function mercury_management_rediscover_host_log(&$uow, $str)
{
  mercury_management_rediscover_log($uow['log_fh'], "host " . $uow['host_name'] . ": " . $str);
}


/**
 *  Callback in the event of a select timeout while processing a host UoW.
 */
function mercury_management_rediscover_host_timeout_cb(&$uow, $idx)
{
  $uow['error'] = 'command timeout!';
  
  mercury_management_rediscover_host_log($uow, "select timeout declared at idx $idx! command follows:\n");
  mercury_management_rediscover_host_log($uow, $uow['cmds'][$idx]);
  
  $host = $uow['host_name'];
  mercury_utilities_fmc_log_message('Host Refresh', t('Timeout accessing host @host. Address host or infrastructure issue and try again.', array('@host' => $host)), WATCHDOG_ERROR);

  # To do: add error handling for sfcdb lockup here.
  
  # Return true to allow the rest of the discovery to complete.  This UoW has 
  # failed.
  return true;
}


/**
 * Discovery+Refresh (Host) Process
 * 
 * Discovery of an individual host is functionally equivalent to performing a 
 * Refresh operation on that host. The process proceeds in two phases, as 
 * follows:
 * 
 * Phase A (Discovery)
 * 1. Gather Host-level information, including whether the FA agent is installed. 
 *    There are several potential error conditions that might occur here and will 
 *    cause Phase A to be terminated prematurely. These include: the host is found 
 *    not to exist; the host is powered down; the host cannot be queried for any 
 *    reason. 
 * 2. If the FA host agent is NOT installed on this host, go to Phase B with data 
 *    level HD ("Host Data"). Otherwise proceed to the next step. 
 * 3. Obtain information from the FA host agent about the current cache device 
 *    (if any) and the cache state. 
 * 4. If there is a current cache device, go to Phase B with data level HD+AD 
 *    ("Agent Data"). Otherwise proceed to the next step. 
 * 5. Obtain a list of cache-capable devices attached to this host (that is, 
 *    devices that could be claimed by FA for caching purposes, but currently are 
 *    not). Go to Phase B with data level HD+AD+DD ("Device Data"). 
 * NOTE: if Phase A is terminated prematurely due to an error condition, an 
 * appropriate error message will be provided to the user, and Phase B will be 
 * skipped.
 * 
 * Phase B (Database Update)
 * 1. Upon entry to Phase B, there are three possible data levels: HD, HD+AD, 
 *    or HD+AD+DD (ordered from "low" to "high"). 
 *    a) Delete any database nodes associated with this host that contain data for 
 *       a higher data level than the current level. For example, if Phase B was 
 *       entered at data level HD, delete any database nodes that contain only AD 
 *       or DD data.
 *    b) If there are nodes that contain mixtures of HD, AD, and DD data fields, 
 *       reinitialize all fields whose data level is higher than the current level. 
 *    c) Delete any database nodes associated with this host that should no longer 
 *       exist given the currently available data. For example, if the data level is 
 *       HD+AD+DD and cache-capable devices are stored in their own nodes, delete 
 *       any nodes representing cache-capable devices on the current host that do not 
 *       appear in the current list of cache-capable devices. 
 * 2. Create and datafill, or load and update, all nodes necessary to store the 
 *    current data level for this host. 
 * 3. Save all created or updated nodes in the database. 
 */

/**
 *  Assemble the discovery UoW for the given host.
 */
function mercury_management_assemble_host_uow($log_fh, $vc_hostname, $vc_username, $vc_password, $host_name, $datacenter_nid, $is_annotated, $is_installing) 
{
  $uow['type'] = 'disc_host';

  $uow['cmds'] = array();
  $uow['cbks'] = array();

  $uow['cmds'][] = mercury_management_wcmd_hostinfo($host_name);
  $uow['cbks'][] = 'mercury_management_wa_cb_disc_host_hostinfo';

  $uow['vc_hostname']    = $vc_hostname;
  $uow['vc_username']    = $vc_username;
  $uow['vc_password']    = $vc_password;
  $uow['host_name']      = $host_name;
  $uow['datacenter_nid'] = $datacenter_nid;
  $uow['is_annotated']   = $is_annotated;
  $uow['is_installing']  = $is_installing;
  $uow['log_fh']         = $log_fh;
  $uow['timeout_cbk']    = 'mercury_management_rediscover_host_timeout_cb';
  
  mercury_management_rediscover_host_log($uow, "obtaining hostinfo\n");
  return $uow;
}


/**
 *  Discovery succeeded; execute Phase B and terminate uow.
 */
function mercury_management_host_discovery_successful(&$uow, $idx, $data_level)
{
   mercury_management_discover_host_db_commit($uow, $data_level);
   mercury_management_rediscover_host_log($uow, "discovery completed successfully for this host\n");
   $uow['cmds'][$idx + 1] = "all_done";
}


/**
 *  Next discovery step: hostinfo now available.
 */
function mercury_management_wa_cb_disc_host_hostinfo(&$uow, $idx)
{
  $result = $uow['results'][$idx];
  $host_info_with_key = mercury_management_parse_hostinfo($result);
  if (empty($host_info_with_key)) {
    $uow['error'] = "couldn't parse hostinfo!";
    mercury_management_rediscover_host_log($uow, $uow['error'] . "\n");
    return false;
  }
  
  mercury_management_rediscover_host_log($uow, "creating/updating host node\n");
  $datacenter_nid = $uow['datacenter_nid'];
  $host_info = mercury_management_create_or_update_host_node($uow, $host_info_with_key, $datacenter_nid);
  $uow['host_info'] = $host_info;

  // Check for pending alarms
  if (mercury_management_abort_updatehost_due_to_pending_alarms($uow, $host_info_with_key)) {
    $uow['error'] = "encountered pending alarms!";
    mercury_management_rediscover_host_log($uow, $uow['error'] . "\n");
    return false;
  }
  
  // Next step is to get the status of our FA VIBs
  mercury_management_rediscover_host_log($uow, "obtaining FA VIB status\n");
  $host_name = $host_info->title;
  $host_username = mercury_management_get_host_username($host_info);
  $host_password = mercury_management_get_host_password($host_info);
  if ($uow['is_annotated'] == false) {
    $uow['cmds'][$idx + 1] = mercury_management_wcmd_host_fa_vib_status($host_name, $host_username, $host_password);
  } else {
    $uow['results'][$idx + 1][0] = FAMC_HOST_BUNDLE_STATUS_INSTALLED;
    $uow['cmds'][$idx + 1] = "no_op";
  }

  $uow['cbks'][$idx + 1] = "mercury_management_wa_cb_disc_host_checkforhostbundle";
  
  return true;
}


/**
 *  Next discovery step: host software "bundle_status" now available.
 */
function mercury_management_wa_cb_disc_host_checkforhostbundle(&$uow, $idx) 
{
  $bundle_status = $uow['results'][$idx][0];
  $uow['bundle_status'] = $bundle_status;
  $host_info = $uow['host_info'];
  $host_name = $uow['host_name'];
  
  mercury_management_rediscover_host_log($uow, "bundle status is $bundle_status\n");
   
  $host_info->field_bundle_status['und'][0]['value'] = $bundle_status;

  if ($uow['is_annotated'] == false && $bundle_status == FAMC_HOST_BUNDLE_STATUS_INSTALLED) {
    mercury_management_access_annotations('host', trim($host_info->title), 'set');
  } else if ($bundle_status == FAMC_HOST_BUNDLE_STATUS_NOT_INSTALLED) {
    mercury_management_access_annotations('host', trim($host_info->title), 'clear');
  }
  
  if (mercury_management_abort_updatehost_due_to_bundle_status($uow, $bundle_status)) {
    $uow['error'] = "couldn't determine VIB status!";
    mercury_management_rediscover_host_log($uow, $uow['error'] . "\n");
    return false;
  }

  $vcInfo = mercury_management_vcenter_details();
  $rediscover = $vcInfo->field_rediscover_status['und'][0]['value'];
  $is_installing = $uow['is_installing'];

  if ((FAMC_HOST_BUNDLE_STATUS_INSTALLED != $bundle_status) || (true == $is_installing)) {
    mercury_management_host_discovery_successful($uow, $idx, 'HD');
  } else {
    mercury_management_rediscover_host_log($uow, "getting more host data with host_control.py 1\n");
    $host_name = $uow['host_name'];
    $host_info = $uow['host_info'];
    $host_username = mercury_management_get_host_username($host_info);
    $host_password = mercury_management_get_host_password($host_info);
    $uow['cmds'][$idx + 1] = mercury_management_wcmd_host_control_1($host_name, $host_username, $host_password);
    $uow['cbks'][$idx + 1] = "mercury_management_wa_cb_disc_host_update_host_config";
  }
  
  return true;
}


/**
 *  Next discovery step: host_control.py 1 output now available.
 */
function mercury_management_wa_cb_disc_host_update_host_config(&$uow, $idx) 
{
  $host_info = $uow['host_info'];
  $vc_hostname = $uow['vc_hostname'];
  $vc_username = $uow['vc_username'];
  $vc_password = $uow['vc_password'];
  $host_name = $uow['host_name'];
  $host_username = mercury_management_get_host_username($host_info);
  $host_password = mercury_management_get_host_password($host_info);
  $result = $uow['results'][$idx];

  # Normally we don't hit this error, so not investing in parallelizing for now
  if (count($result) == 0) {
    if ($uow['is_annotated'] == true) {
    # Accounting for the corner case wherein the annotation may be falsely present
      $uow['is_annotated'] = false;
      mercury_management_rediscover_host_log($uow, "obtaining FA VIB status\n");
      $uow['cmds'][$idx + 1] = mercury_management_wcmd_host_fa_vib_status($host_name, $host_username, $host_password);
      $uow['cbks'][$idx + 1] = "mercury_management_wa_cb_disc_host_checkforhostbundle";
      return true;
    }
      
    mercury_management_rediscover_host_log($uow, "attempting sfcbd fix\n");
    $result = mercury_management_attempt_sfcbd_fix($vc_hostname, $vc_username, $vc_password, $host_name, $host_username, $host_password);
    if (count($result) == 0) {
      $uow['error'] = "sfcbd fix failed!";
      mercury_management_rediscover_host_log($uow, $uow['error'] . "\n");
      return false;
    }
  }

  mercury_management_rediscover_host_log($uow, "updating host node with host_control_1 data\n");  
  $host_info_with_key = mercury_management_parse_host_control_1($result);
  mercury_management_add_host_control_1_data_to_host_node($host_info_with_key, $host_info);
  
  if ($host_info->field_naa_identifier['und'][0]['value'] == '') {
    mercury_management_rediscover_host_log($uow, "obtaining host device list\n");
    $uow['cmds'][$idx + 1] = mercury_management_wcmd_esxcli_device_list($vc_hostname, $vc_username, $vc_password, $host_name);
    $uow['cbks'][$idx + 1] = "mercury_management_wa_cb_disc_host_get_device_array_got_device_list";
  } else {
    mercury_management_host_discovery_successful($uow, $idx, 'HD+AD');
  }
  
  return true;
}


/**
 *  Next discovery step: host device list now available.
 */
function mercury_management_wa_cb_disc_host_get_device_array_got_device_list(&$uow, $idx)
{
  $uow['device_list'] = $uow['results'][$idx];
  mercury_management_rediscover_host_log($uow, "obtaining host path list\n");
  $vc_hostname = $uow['vc_hostname'];
  $vc_username = $uow['vc_username'];
  $vc_password = $uow['vc_password'];
  $host_name = $uow['host_name'];
  $uow['cmds'][$idx + 1] = mercury_management_wcmd_esxcli_path_list($vc_hostname, $vc_username, $vc_password, $host_name);
  $uow['cbks'][$idx + 1] = "mm_discover_host_wa_cb_disc_host_coredump_partition_list";
  return true;
}


/**
 * Next discovery step: Get diagnostic devices that aren't supposed to be claimed
 */
function mm_discover_host_wa_cb_disc_host_coredump_partition_list(&$uow, $idx)
{
  $uow['path_list'] = $uow['results'][$idx];
  mercury_management_rediscover_host_log($uow, "obtaining host system coredump partition list\n");
  $vc_hostname = $uow['vc_hostname'];
  $vc_username = $uow['vc_username'];
  $vc_password = $uow['vc_password'];
  $host_name = $uow['host_name'];
  $uow['cmds'][$idx + 1] = mercury_management_wcmd_esxcli_system_coredump_partition_list($vc_hostname, $vc_username, $vc_password, $host_name);
  $uow['cbks'][$idx + 1] = "mercury_management_wa_cb_disc_host_get_device_array";
  return true;
}


/**
 *  Next discovery step: host path list now available.
 */
function mercury_management_wa_cb_disc_host_get_device_array(&$uow, $idx)
{
  $dataArr = $uow['device_list'];
  $dataPathArr = $uow['path_list'];
  $diagnosticPartitionArr = $uow['results'][$idx];
  
  mercury_management_rediscover_host_log($uow, "parsing host device and path lists\n");
  $deviceArray = mercury_management_parse_esx_device_and_path_lists($dataArr, $dataPathArr, $diagnosticPartitionArr);
  $uow['device_array'] = $deviceArray;
  mercury_management_host_discovery_successful($uow, $idx, 'HD+AD+DD');
  return true;
}


/**
 *  Perform Phase B (Database Update) for the given host
 */
function mercury_management_discover_host_db_commit(&$uow, $data_level)
{
   /*
    * Differences between HD and HD+AD are dealt with by 
    * mercury_management_add_host_control_1_data_to_host_node; all we have 
    * to do is create the host node, and create/update or delete the device
    * nodes.
    */

   mercury_management_rediscover_host_log($uow, "committing host to database ($data_level)\n");
   $host_info = $uow['host_info'];
   node_save($host_info);
   if ($data_level == 'HD+AD+DD') {
      mercury_management_rediscover_host_log($uow, "committing host device nodes to database\n");
      $deviceArray = $uow['device_array'];
      mercury_management_update_host_device_db_commit($uow, $deviceArray, $host_info->nid);
   } else {
      if (false == $uow['host_info_is_new']) {
         mercury_management_rediscover_host_log($uow, "removing host device nodes from database\n");
         mercury_management_update_host_device_db_commit($uow, array(), $host_info->nid);
      }
   }
}
