<?php
/**
 *  Flash Accel Management Console (FMC)
 *  Copyright (C) 2013, NetApp, Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Javascript
 *
 *    Farbtastic - Copyright (c) 2010 Matt Farina
 *
 *    jQuery - Copyright (c) 2010 John Resig
 * 
 *    jQuery BBQ - Copyright (c) 2010 "Cowboy" Ben Alman
 *  
 *    jQuery Cookie - Copyright (c) 2006 Klaus Hartl
 *  
 *    jQuery Form - Copyright (c) 2010 Mike Alsup
 *  
 *    jQuery Once - Copyright (c) 2009 Konstantin K\ufffdfer
 *
 *    jQuery UI - Copyright (c) 2010 by the original authors
 *      (http://jqueryui.com/about)
 *
 *    Sizzle.js - Copyright (c) 2010 The Dojo Foundation (http://sizzlejs.com/)
 *
 *  PHP
 *
 *    ArchiveTar - Copyright (c) 1997 - 2008 Vincent Blavet
 */

define('FAMC_VM_DRIVER_STATUS_INSTALLED',      'Agent Installed');
define('FAMC_VM_DRIVER_STATUS_NOT_INSTALLED',  'Agent Not Installed');
define('FAMC_VM_DRIVER_STATUS_NOT_COMPATIBLE', 'Not Compatible');
define('FAMC_VM_DRIVER_STATUS_NOT_KNOWN',      'Unable to Communicate with VM');
define('FAMC_WEBSERVICE_DEFAULT_PORT',         '7010');


/**
 *  Create or update the database entry (node) for a given VM.
 */
function mercury_management_create_or_update_vm_node(&$uow, $vm_info_with_key, $host_nid)
{
  global $user;
 
  $action = "<undefined>";
  $nid = db_query("SELECT nid FROM {node} WHERE type = :tp AND title = :ttl", array(':tp' => 'virtual_machine', ':ttl' => trim($vm_info_with_key['Name'])))->fetchField();
  if (!empty($nid)) {
    $vm_info = node_load($nid);
    $uow['vm_info_orig'] = clone $vm_info;
    $action = "updated";
  }
  else {
    $vm_info = new stdClass();
    $vm_info->type = 'virtual_machine';
    $vm_info->author = $user->name;
    $vm_info->status = 1;
    $vm_info->uid = $user->uid;
    $vm_info->comment = 0;
    $vm_info->language = 'und';
    $vm_info->format = 2;
    $vm_info->title = $vm_info_with_key['Name'];
    $action = "created";
  }
  
  mercury_management_initialize_vm_node($vm_info);
  
  $vm_info->field_vm_ip_address['und'][0]['value'] = $vm_info_with_key['IP Address'];
  $vm_info->field_no_of_cpu['und'][0]['value'] = $vm_info_with_key['No. of CPU(s)'];
  $vm_info->field_memory_size['und'][0]['value'] = $vm_info_with_key['Memory Size'];
  $vm_info->field_virtual_disks['und'][0]['value'] = $vm_info_with_key['Virtual Disks'];
  $vm_info->field_template['und'][0]['value'] = $vm_info_with_key['Template'];
  $vm_info->field_vm_path_name['und'][0]['value'] = $vm_info_with_key['vmPathName'];
  $vm_info->field_guest_os['und'][0]['value'] = $vm_info_with_key['Guest OS'];
  $vm_info->field_guest_id['und'][0]['value'] = $vm_info_with_key['guestId'];
  $vm_info->field_vm_hostname['und'][0]['value'] = $vm_info_with_key['Host name'];
  $vm_info->field_cpu_usage['und'][0]['value'] = $vm_info_with_key['Cpu usage'];
  $vm_info->field_host_memory_usage['und'][0]['value'] = $vm_info_with_key['Host memory usage'];
  $vm_info->field_guest_memory_usage['und'][0]['value'] = $vm_info_with_key['Guest memory usage'];
  $vm_info->field_host['und'][0]['target_id'] = $host_nid;
  if (array_key_exists('VMware Tools', $vm_info_with_key)) {
    $vm_info->field_vmware_tools['und'][0]['value'] = $vm_info_with_key['VMware Tools'];
  }
  node_save($vm_info);
  
  $uow['vm_info'] = $vm_info;
  if ($action == 'updated') {
    $uow['vm_info_is_new'] = false;
  } else {
    $uow['vm_info_is_new'] = true; 
  }

  $nid = $vm_info->nid;
  mercury_management_rediscover_vm_log($uow, "$action vm nid $nid (host_nid $host_nid)\n");
  
  return $vm_info;
}


/**
 * Update the driver_status field on a phase A failure.  If the VM is not in 
 * the database, create a node for it.  Otherwise use the original node values
 * except for driver_status, so we can be sure not to leave the database in
 * an inconsistent state.
 */
function mercury_management_update_driver_status_on_phase_a_failure(&$uow, $status)
{
   if (false == $uow['vm_info_is_new']) {
      $vm_info = $uow['vm_info_orig'];
   } else {
      $vm_info = $uow['vm_info'];
   }
   $vm_info->field_driver_status['und'][0]['value'] = $status;
   node_save($vm_info);
}


/**
 * Add cache disk to VM database node. 
 */
function mercury_management_add_cache_disk_to_vm_node(&$vm_info, $vDiskArr)
{
  $cacheDevice = '';
  if (!empty($vDiskArr)) {
    foreach ($vDiskArr as $key => $vDisk) {
      if ($vDisk['type'] == "CACHE") {
        $vm_info->field_cachespace['und'][0]['value'] = $vDisk['id'];
        $vm_info->field_fa_cache_size['und'][0]['value'] = $vDisk['Blocks'];
        $vm_info->field_fa_sector_size['und'][0]['value'] = $vDisk['Blocksize'];
        $cacheDevice = $vDisk['id'];
        break;
      }
    }
  }
  return $cacheDevice;
}


/**
 *  Create or update the vdisk database entries associated with a VM.
 */
function mercury_management_update_vm_vdisk_db_commit(&$uow, $attrInfo, $vm_nid)
{
   global $user;

   $vdisk_nids = mercury_management_find_dependent_nodes(array($vm_nid), 'vdisk', 'field_virtual_machine');
   $nvn = count($vdisk_nids);

   # Optimize the case where we're deleting all vdisk nodes.
   if (empty($attrInfo) || ($attrInfo['disk_count'] == 0)) {
      if ($nvn) {
         mercury_management_delete_vdisks($uow['log_fh'], $vdisk_nids);
      } else {
         mercury_management_rediscover_vm_log($uow, "couldn't find any vdisk nodes linked to vm_nid $vm_nid\n");
      }
      return true;
   }

   # We have some vdisks, some of which may already have nodes.

   # Load all existing nodes; we're going to delete the ones that don't get
   # updated.      
   if (isset($vdisk_nids)) {
      $vdisk_nodes = entity_load('node', $vdisk_nids);
   } else {
      $vdisk_nodes = array();
   }

   # Search for each of the passed-in vdisks in the existing vdisk nodes.
   $diskCount = $attrInfo['disk_count'];
   $vDiskArr = $uow['vDiskArr'];
   for ($index = 0; $index < $diskCount; $index++) {
      $identifier = "naa." . $attrInfo[$index]['vdisk_identifier'];
      $identifier = strtolower(trim($identifier));
      $action = "creating";
      foreach ($vdisk_nodes as $vi_key => $vdisk_info) {
         if ($vdisk_info->title === $identifier) {
             $action = "updating";
             break;
         }
      }
      if ($action == "creating") {
         $vdisk_info = new stdClass();
         $vdisk_info->type = 'vdisk';
         $vdisk_info->author = $user->name;
         $vdisk_info->status = 1;
         $vdisk_info->uid = $user->uid;
         $vdisk_info->comment = 0;
         $vdisk_info->language = 'und';
         $vdisk_info->format = 2;
         $vdisk_info->title = $identifier;
      } else {
         # Later we're going to delete everything left in vdisk_nodes
         unset($vdisk_nodes[$vi_key]);
      }
      mercury_management_initialize_vdisk_node($vdisk_info);

      $vdisk_info->field_virtual_machine['und'][0]['target_id'] = $vm_nid;
      //Figure out the appropriate type for the identifier
    
      $disk_type='Direct';
      if (!empty($vDiskArr)) {
        foreach ($vDiskArr as $key => $value){
         // watchdog('test', $identifier .'='. $value['id']);
          if (strtolower($identifier)== strtolower($value['id'])) {
            $disk_type=$value['type'];
          }
        }
      }
      $vdisk_info->field_vdisk_type['und'][0]['value'] = $disk_type;
      $vdisk_info->field_vdisk_name['und'][0]['value'] = '';
      $vdisk_info->field_capacity['und'][0]['value'] = 0;
      $vdisk_info->field_freespace['und'][0]['value'] = 0;

      # Update the cache information
      $vdisk_info->field_fa_cache_mode['und'][0]['value'] = $attrInfo[$index]['cache_mode'];
      $vdisk_info->field_fa_administration_state['und'][0]['value'] = $attrInfo[$index]['admin_state'];
      $vdisk_info->field_fa_operational_state['und'][0]['value'] = $attrInfo[$index]['op_state'];
      $vdisk_info->field_fa_operational_message['und'][0]['value'] = $attrInfo[$index]['op_message'];
      $vdisk_info->field_fa_performance_state['und'][0]['value'] = $attrInfo[$index]['perf_state'];
      $vdisk_info->field_vdisk_identifier['und'][0]['value'] = $attrInfo[$index]['vdisk_identifier'];
      if ($attrInfo[$index]['path'] != '') {
         $vdisk_info->field_path['und'][0]['value'] = $attrInfo[$index]['path'];
      }

      $vdisk_info->field_disk_index['und'][0]['value'] = $index;
      $vdisk_info->field_vdisk_id['und'][0]['value'] = "naa." . $attrInfo[$index]['vdisk_identifier'];
            
      node_save($vdisk_info);
      
      $nid = $vdisk_info->nid;
      mercury_management_rediscover_vm_log($uow, "$action vdisk nid $nid (vm_nid $vm_nid) id $identifier\n");
   }

   # Any device still "set" in vdisk_nodes needs to be deleted   
   $vdisk_nids_to_delete = array();
   foreach ($vdisk_nodes as $vdisk_info) {
      $nid = $vdisk_info->nid;
      $id = $vdisk_info->title;
      mercury_management_rediscover_vm_log($uow, "deleting vdisk nid $nid (vm_nid $vm_nid) id $id\n");
      $vdisk_nids_to_delete[] = $vdisk_info->nid;
   }
   if (!empty($vdisk_nids_to_delete)) {
      mercury_management_delete_vdisks($uow['log_fh'], $vdisk_nids_to_delete);
   }

   return true;
}


/**
 *  Obtain the VM hostname from the VM node data.
 */
function mercury_management_get_vm_hostname($vm_info)
{
  if ($vm_info->field_vm_hostname) {
    $vm_hostname = $vm_info->field_vm_hostname['und'][0]['value'];
  } else {
    $vm_hostname = "";
  }
  return $vm_hostname;
}


/**
 *  Obtain the VM webservice port from the VM node data.
 */
function mercury_management_get_vm_port($vm_info)
{
  if (!empty($vm_info->field_install_port)) {
    $port = $vm_info->field_install_port['und'][0]['value'];
  } else {
    $port = FAMC_WEBSERVICE_DEFAULT_PORT;
  }
  return $port;
}


/**
 *  Decide whether VM discovery should be aborted due to an unknown VM hostname.
 */
function mercury_management_abort_updatevm_due_to_unknown_vm_hostname(&$uow)
{
  $vm_info = $uow['vm_info'];
  $vm_hostname = mercury_management_get_vm_hostname($vm_info);

  if (empty($vm_hostname) || ($vm_hostname == 'Not Known')) {
    mercury_utilities_fmc_log_message('VM Refresh', t('Cannot connect to VM @vm. VM may be powered off.', array('@vm' => $vm_info->title)), WATCHDOG_WARNING);
    mercury_management_update_driver_status_on_phase_a_failure($uow, FAMC_VM_DRIVER_STATUS_NOT_KNOWN);
    # Peg an actual discovery error
    $uow['error'] = "unknown vm hostname!";
    mercury_management_rediscover_vm_log($uow, "discovery terminated: " . $uow['error'] . "\n");
    return true;
  }
  
  return false;
}


/**
 *  Decide whether VM discovery should be aborted due to an unresolvable VM hostname.
 */
function mercury_management_abort_updatevm_due_to_unresolvable_vm_hostname(&$uow, $result)
{
  $vm_info = $uow['vm_info'];
  $ip = "0.0.0.0";
  $valid_fqdn = mercury_management_parse_wcmd_validate_fqdn($result, $ip);
  $vm_hostname = mercury_management_get_vm_hostname($vm_info);
  if ($valid_fqdn != true) {
    mercury_utilities_fmc_log_message('VM Refresh', t('VM @vm hostname "@vm_hostname" cannot be resolved from FMC.  Please correct your DNS configuration, and try again.',  array('@vm' => $vm_info->title, '@vm_hostname' => $vm_hostname)), WATCHDOG_WARNING);
    mercury_management_update_driver_status_on_phase_a_failure($uow, FAMC_VM_DRIVER_STATUS_NOT_KNOWN);
    # Peg an actual discovery error
    $uow['error'] = "couldn't resolve vm hostname!";
    mercury_management_rediscover_vm_log($uow, "discovery terminated: " . $uow['error'] . "\n");
    return true;
  }
  
  return false;
}


/**
 *  Decide whether VM discovery should be aborted because the VM is powered off.
 */
function mercury_management_abort_updatevm_due_to_vm_powered_off(&$uow, $vm_info_with_key)
{
  $vm_info = $uow['vm_info'];
  $powerstate = $vm_info_with_key['Power State'];
  if ($powerstate != "poweredOn") {
    mercury_utilities_fmc_log_message('VM Refresh', t('Cannot connect to VM @vm. VM is powered off.', array('@vm' => $vm_info->title)), WATCHDOG_NOTICE);
    mercury_management_update_driver_status_on_phase_a_failure($uow, FAMC_VM_DRIVER_STATUS_NOT_KNOWN);
    # Do not peg a discovery error
    mercury_management_rediscover_vm_log($uow, "discovery terminated: vm is powered off\n");
    return true;
  }

  return false;
}


/**
 *  Decide whether a given string corresponds to a supported guest OS.
 */
function mercury_management_guest_os_is_supported($guest_os)
{
  if ($guest_os == 'Microsoft Windows Server 2008 R2 (64-bit)') {
    return true;
  }
  
  return false;
}


/**
 *  Decide whether VM discovery should be aborted due to an unsupported guest OS.
 */
function mercury_management_abort_updatevm_due_to_unsupported_guest_os(&$uow)
{
  $vm_info = $uow['vm_info'];
  $guest_os = $vm_info->field_guest_os['und'][0]['value'];
  if (false == mercury_management_guest_os_is_supported($guest_os)) {
    if ($guest_os == 'Not Known') {
      mercury_utilities_fmc_log_message('VM Refresh', t('VM @vm is running an unknown guest OS. VM may be shut down or rebooting.', array('@vm' => $vm_info->title)), WATCHDOG_NOTICE);
      mercury_management_update_driver_status_on_phase_a_failure($uow, FAMC_VM_DRIVER_STATUS_NOT_KNOWN);
      # Do not peg a discovery error
      mercury_management_rediscover_vm_log($uow, "discovery terminated: guest os not known\n");
    } else {
      mercury_utilities_fmc_log_message('VM Refresh', t('VM @vm is running an unsupported guest OS.', array('@vm' => $vm_info->title)), WATCHDOG_NOTICE);
      mercury_management_update_driver_status_on_phase_a_failure($uow, FAMC_VM_DRIVER_STATUS_NOT_COMPATIBLE);
      # Do not peg a discovery error
      mercury_management_rediscover_vm_log($uow, "discovery terminated: unsupported guest os $guest_os\n");
    }
    return true; 
  }
  return false;
}


/**
 *  Decide whether VM discovery should be aborted due to login failure.
 */
function mercury_management_abort_updatevm_due_to_bad_credentials(&$uow, $validate_username_password)
{
  if ($validate_username_password != "Success") {
    $vm_info = $uow['vm_info'];
    if ($validate_username_password == "Error") {
      mercury_utilities_fmc_log_message('VM Refresh', t('Username/Password incorrect for VM @vm. Please set the correct username/password and try again', array('@vm' => $vm_info->title)), WATCHDOG_ERROR);
      mercury_management_update_driver_status_on_phase_a_failure($uow, FAMC_VM_DRIVER_STATUS_NOT_KNOWN);
    } else if ($validate_username_password == "CannotConnect") {
      mercury_utilities_fmc_log_message('VM Refresh', t('Cannot connect to VM @vm. Please check if it is powered on and try again.', array('@vm' => $vm_info->title)), WATCHDOG_WARNING);
      mercury_management_update_driver_status_on_phase_a_failure($uow, FAMC_VM_DRIVER_STATUS_NOT_KNOWN);
    }
    # Peg an actual discovery error
    $uow['error'] = "invalid vm credentials!";
    mercury_management_rediscover_vm_log($uow, "discovery terminated: " . $uow['error'] . "\n");
    return true;
  }
  
  return false;
}


/**
 *  Decide whether VM discovery should be aborted due to missing VMware Tools.
 */
function mercury_management_abort_updatevm_due_to_no_vmware_tools(&$uow)
{
  $vm_info = $uow['vm_info'];
  $driver_status = "OK";
  if (!empty($vm_info->field_vmware_tools)) {
    if ($vm_info->field_vmware_tools['und'][0]['value'] == 'VMware Tools is not running.' ||
	$vm_info->field_vmware_tools['und'][0]['value'] == 'VMware Tools has never been installed or has not run in the virtual machine.' || 
	$vm_info->field_vm_hostname['und'][0]['value'] == '') {
      $driver_status = FAMC_VM_DRIVER_STATUS_NOT_KNOWN;
    }
    # Do not peg a discovery error
  } else {
    $driver_status = FAMC_VM_DRIVER_STATUS_NOT_COMPATIBLE;
    # Peg a discovery error
    $uow['error'] = "no vmware tools!";
  }
  
  if ($driver_status != "OK") {
    mercury_utilities_fmc_log_message('VM Refresh', t('Could not find VMware Tools on VM @vm.', array('@vm' => $vm_info->title)), WATCHDOG_NOTICE);
    mercury_management_update_driver_status_on_phase_a_failure($uow, $driver_status);
    mercury_management_rediscover_vm_log($uow, "discovery terminated: no vmware tools ($driver_status)\n");
    return true;
  }

  return false;
}


/**
 *  Set driver status to Agent Installed
 */
function mercury_management_updatevm_set_driver_status_installed(&$uow)
{
  $vm_info = $uow['vm_info'];
  $vm_info->field_driver_status['und'][0]['value'] = FAMC_VM_DRIVER_STATUS_INSTALLED;
}


/**
 *  Decide whether VM discovery should be aborted due to bad driver status
 */
function mercury_management_abort_updatevm_due_to_driver_status(&$uow, $driver_status)
{
  $vm_info = $uow['vm_info'];
  $vm_info->field_driver_status['und'][0]['value'] = $driver_status;

  if (($driver_status != FAMC_VM_DRIVER_STATUS_INSTALLED) &&
      ($driver_status != FAMC_VM_DRIVER_STATUS_NOT_INSTALLED)) {
    mercury_utilities_fmc_log_message('VM Refresh', t('Bad driver status on on VM @vm.', array('@vm' => $vm_info->title)), WATCHDOG_WARNING);
    mercury_management_update_driver_status_on_phase_a_failure($uow, $driver_status);
    # Peg a discovery error
    $uow['error'] = "bad driver status!";
    mercury_management_rediscover_vm_log($uow, "discovery terminated: " . $uow['error'] . "\n");
    return true;
  }
  
  return false;
}


/**
 *  Log a string to the log file associated with this UoW, tagged with VM name.
 */
function mercury_management_rediscover_vm_log(&$uow, $str)
{
  mercury_management_rediscover_log($uow['log_fh'], "vm " . $uow['vm'] . ": " . $str);
}


/**
 *  Callback in the event of a select timeout while processing a VM UoW.
 */
function mercury_management_rediscover_vm_timeout_cb(&$uow, $idx)
{
  $uow['error'] = 'command timeout!';
  mercury_management_rediscover_vm_log($uow, "select timeout declared at idx $idx! command follows:\n");
  mercury_management_rediscover_vm_log($uow, $uow['cmds'][$idx]);

  $vm = $uow['vm'];
  mercury_utilities_fmc_log_message('VM Refresh', t('Timeout accessing VM @vm. Address VM or infrastructure issue and try again.', array('@vm' => $vm)), WATCHDOG_ERROR);
  
  # Return true to allow the rest of the discovery to complete.  This UoW has 
  # failed.
  return true;
}


/**
 * Discovery+Refresh (VM) Process
 * 
 * Discovery of an individual VM is functionally equivalent to performing a 
 * refresh operation on that VM. The process proceeds in two phases, as 
 * follows:
 * 
 * Phase A (Discovery)
 * 1. Gather VM-level information, including whether the FA agent is installed. 
 *    There are several potential error conditions that might occur here and 
 *    will cause Phase A to be terminated prematurely. These include: the VM 
 *    is found not to exist; the VM is powered down; the VM cannot be queried 
 *    for any reason. 
 * 2. If the FA VM agent is NOT installed on this VM, go to Phase B with data 
 *    level VD ("VM Data"). Otherwise proceed to the next step. 
 * 3. Obtain information from the FA VM agent about the current cache 
 *    configuration. If cache space is not allocated to this VM, go to Phase B 
 *    with data level VD+CD ("Cache Data"). Otherwise proceed to the next step. 
 * 4. Obtain information about the vDisks attached to this VM. Go to Phase B 
 *    with data level VD+CD+DD ("Disk Data"). Implementation note: based on the 
 *    Barium UI Functional Spec, saving vDisk data in the FMC database may no 
 *    longer be necessary. If so, this step will be omitted. 
 * NOTE: if Phase A is terminated prematurely due to an error condition, an 
 * appropriate error message will be provided to the user, and Phase B will be 
 * skipped.
 * 
 * Phase B (Database Update)
 * 1. Upon entry to Phase B, there are three possible data levels: VD, VD+CD, 
 *    or VD+CD+DD (ordered from "low" to "high"). 
 *    a) Delete any database nodes associated with this VM that contain data 
 *       for a higher data level than the current level. For example, if Phase 
 *       B was entered at data level VD, delete any database nodes that contain
 *       only CD or DD data. 
 *    b) If there are nodes that contain mixtures of VD, CD, and DD data fields,
 *       reinitialize all fields whose data level is higher than the current 
 *       level. 
 *    c) Delete any database nodes associated with this VM that should no longer 
 *       exist given the currently available data. For example, if the data l
 *       evel is VD+CD+DD and vDisks are stored in their own nodes, delete any 
 *       nodes representing vDisks on the current VM that do not appear in the 
 *       current list of vDisks. 
 * 2. Create and datafill, or load and update, all nodes necessary to store the 
 *    current data level for this VM. 
 * 3. Save all created or updated nodes in the database. 
 */


/**
 *  Assemble the discovery UoW for the given VM.
 */
function mercury_management_assemble_vm_uow($log_fh, $vc_hostname, $vc_username, $vc_password, $vm, $is_annotated, $host_nid) 
{
  $uow['type'] = 'disc_vm';

  $uow['cmds'] = array();
  $uow['cbks'] = array();

  $uow['cmds'][] = mercury_management_wcmd_vminfo($vm);
  $uow['cbks'][] = 'mercury_management_wa_cb_disc_vm_vminfo';

  $uow['vc_hostname'] = $vc_hostname;
  $uow['vc_username'] = $vc_username;
  $uow['vc_password'] = $vc_password;
  $uow['vm']          = $vm;
  $uow['host_nid']    = $host_nid;
  $uow['log_fh']      = $log_fh;
  $uow['is_annotated']= $is_annotated;
  $uow['timeout_cbk'] = 'mercury_management_rediscover_vm_timeout_cb';
  
  mercury_management_rediscover_vm_log($uow, "obtaining vminfo\n");

  return $uow;
}


/**
 *  Discovery succeeded; execute Phase B and terminate uow.
 */
function mercury_management_vm_discovery_successful(&$uow, $idx, $data_level)
{
   mercury_management_discover_vm_db_commit($uow, $data_level);
   mercury_management_rediscover_vm_log($uow, "discovery succeeded; vm updated in database\n");
   $uow['cmds'][$idx + 1] = "all_done";
}


/**
 *  Vminfo now available.
 */
function mercury_management_wa_cb_disc_vm_vminfo(&$uow, $idx)
{
  $result = $uow['results'][$idx];
  $vm_info_with_key = mercury_management_parse_vminfo($result);
  
  if (empty($vm_info_with_key)) {
    $uow['error'] = "couldn't parse vminfo!";
    mercury_management_rediscover_vm_log($uow, $uow['error'] . "\n");
    return false;    
  }
  $uow['vm_info_with_key'] = $vm_info_with_key;
  
  mercury_management_rediscover_vm_log($uow, "creating/updating vm node\n");
  
  // We know the current host node. Query vcenter to see if VM is still running on the same host.
  $prev_host_nid = $uow['host_nid'];

  // Query vcenter for VM to Host mapping
  $vcInfo = mercury_management_vcenter_details();
  $vc_hostname = $vcInfo->field_hostname['und'][0]['value'];
  $vc_username = fmc_cipher::decrypt($vcInfo->field_username['und'][0]['value']);
  $vc_password = fmc_cipher::decrypt($vcInfo->field_password['und'][0]['value']);
  $get_vm_info = exec("perl /usr/lib/vmware-vcli/apps/general/vcenterdiscover.pl --url https://" . $vc_hostname . "/sdk/vimService  --username '" . $vc_username . "' --password '" . $vc_password . "' --operation vmlist", $dataArr);
  if (!empty($dataArr)) {
    foreach ($dataArr as $item) {
      $host_split = explode('Host :', $item);
      if (count($host_split) > 1) {
        $currHost= trim($host_split[1]);  
      }
      $vm_split = explode('VM :', $item);
      if (count($vm_split) > 1) {
        if (trim($vm_split[1]) == $vm_info_with_key['Name']) {
          $currently_running_host=  $currHost;
        }
      }
    }
  }
  $host_nid = db_query("SELECT nid FROM {node} WHERE type = :tp AND title = :ttl", array(':tp' => 'host', ':ttl' => trim($currently_running_host)))->fetchField();

  // if a host nid is found build and and save VM otherwise do not build the VM
  if ($host_nid) {

    if ($host_nid != $prev_host_nid) {
      $host = node_load($host_nid);
      mercury_utilities_fmc_log_message('VM Refresh', t('VM @vm is running on host @host', array('@vm' => $vm_info_with_key['Name'], '@host' => $host->title)));
    }
    
    $vm_info = mercury_management_create_or_update_vm_node($uow, $vm_info_with_key, $host_nid);
    $uow['vm_info'] = $vm_info;

    mercury_management_rediscover_vm_log($uow, "obtaining vdiskinfo\n");
    $uow['cmds'][$idx + 1] = mercury_management_wcmd_vdiskinfo($uow['vm']);
    $uow['cbks'][$idx + 1] = "mercury_management_wa_cb_disc_vm_vdiskinfo";
  }
  else {
    mercury_management_rediscover_vm_log($uow,  "Error:  VM doesn't have a valid host_nid...skipping node creation\n");
    return false;
  }

  return true;
}


/**
 *  Next discovery step: vdiskinfo now available.
 */
function mercury_management_wa_cb_disc_vm_vdiskinfo(&$uow, $idx)
{
  $result = $uow['results'][$idx];
  $vm_info = $uow['vm_info'];
  $vDiskArr = mercury_management_parse_vdiskinfo($vm_info->title, $result);
 
  $uow['vDiskArr'] = $vDiskArr;

  $cacheDevice = mercury_management_add_cache_disk_to_vm_node($vm_info, $vDiskArr);
  if (!empty($cacheDevice)) {
    mercury_management_rediscover_vm_log($uow, "updating vm node with cache device $cacheDevice\n");
  }
  $uow['cacheDevice'] = $cacheDevice;

  $vm_info_with_key = $uow['vm_info_with_key'];
  if (mercury_management_abort_updatevm_due_to_vm_powered_off($uow, $vm_info_with_key)) {
    return false;
  }

  if (mercury_management_abort_updatevm_due_to_unsupported_guest_os($uow)) {
    return false;
  }

  if (mercury_management_abort_updatevm_due_to_unknown_vm_hostname($uow)) {
    return false;
  }
    
  $vm_hostname = mercury_management_get_vm_hostname($vm_info);
  mercury_management_rediscover_vm_log($uow, "checking DNS for vm hostname $vm_hostname\n");
  $uow['cmds'][$idx + 1] = mercury_management_wcmd_validate_fqdn($vm_hostname);
  $uow['cbks'][$idx + 1] = "mercury_management_wa_cb_disc_vm_validate_fqdn";
  
  return true;
}


/**
 *  Next discovery step: host output now available.
 */
function mercury_management_wa_cb_disc_vm_validate_fqdn(&$uow, $idx)
{
  $result = $uow['results'][$idx];
  $vm_info = $uow['vm_info'];

  if (mercury_management_abort_updatevm_due_to_unresolvable_vm_hostname($uow, $result)) {
    return false;
  }

  # Speculatively invoke the web service.  If it succeeds, FA is installed,
  # and we can skip the vmruns that check for FA installed.
  mercury_management_rediscover_vm_log($uow, "attempting to obtain cache info\n");
  $hostname = mercury_management_get_vm_hostname($vm_info);
  $username = mercury_management_get_vm_username($vm_info);
  $password = mercury_management_get_vm_password($vm_info);
  $method   = 'HgGetCacheOptions';
  $port     = mercury_management_get_vm_port($vm_info);
  $uow['cmds'][$idx + 1] = mercury_management_wcmd_mm_wsdl($hostname, $username, $password, $method, null, $port);
  $uow['cbks'][$idx + 1] = "mercury_management_wa_cb_disc_vm_getcacheoptions_attempt";
  
  return true;
}


/**
 *  Next discovery step: results of speculative web service call now available 
 */
function mercury_management_wa_cb_disc_vm_getcacheoptions_attempt(&$uow, $idx)
{
  $result = implode("\n", $uow['results'][$idx]);
  eval("\$attrInfo = $result;");

  if ($attrInfo == 'error_wsdl_call') {
    # Web service failed but we don't know why; continue with regular flow
    mercury_management_rediscover_vm_log($uow, "speculative web service call failed\n");
    mercury_management_rediscover_vm_log($uow, "checking vm credentials\n");
    $vm_info = $uow['vm_info'];
    $host_and_vc_of_vm_nids = mercury_management_get_host_dc_vc_of_vm($vm_info->nid);
    $vc_of_vm_node = node_load($host_and_vc_of_vm_nids['vcenter']);
    $vc_hostname = $vc_of_vm_node->field_hostname['und'][0]['value'];
    $vc_username = fmc_cipher::decrypt($vc_of_vm_node->field_username['und'][0]['value']);
    $vc_password = fmc_cipher::decrypt($vc_of_vm_node->field_password['und'][0]['value']);
    $username = mercury_management_get_vm_username($vm_info);
    $password = mercury_management_get_vm_password($vm_info);
    $path = $vm_info->field_vm_path_name['und'][0]['value'];
    $uow['cmds'][$idx + 1] = mercury_management_wcmd_vmrun_chk_cred($vc_hostname, $vc_username, $vc_password, $username, $password, $path);
    $uow['cbks'][$idx + 1] = "mercury_management_wa_cb_disc_vm_vmrun_chk_cred";
    return true;
  } else {
    # Web service call succeeded; FA must be installed.  Short-circuit flow.
    mercury_management_updatevm_set_driver_status_installed($uow);
    if ($uow['is_annotated'] == false) {
      $vm_info = $uow['vm_info'];
      mercury_management_access_annotations('vm', $vm_info->title, 'set');
    }
    return mercury_management_wa_cb_disc_vm_getcacheoptions($uow, $idx);
  }
}


/**
 *  Next discovery step: vmrun output for credential check now available.
 */
function mercury_management_wa_cb_disc_vm_vmrun_chk_cred(&$uow, $idx)
{
  $result = $uow['results'][$idx];
  $vm_info = $uow['vm_info'];
  
  $chk_cred = mercury_management_parse_vmrun_chk_cred($result);
  if (mercury_management_abort_updatevm_due_to_bad_credentials($uow, $chk_cred)) {
    return false;
  }
  
  if (mercury_management_abort_updatevm_due_to_no_vmware_tools($uow)) {
    return false;
  }
  
  $host_and_vc_of_vm_nids = mercury_management_get_host_dc_vc_of_vm($vm_info->nid);
  $vc_of_vm_node = node_load($host_and_vc_of_vm_nids['vcenter']);
  $vc_hostname = $vc_of_vm_node->field_hostname['und'][0]['value'];
  $vc_username = fmc_cipher::decrypt($vc_of_vm_node->field_username['und'][0]['value']);
  $vc_password = fmc_cipher::decrypt($vc_of_vm_node->field_password['und'][0]['value']);
  $username = mercury_management_get_vm_username($vm_info);
  $password = mercury_management_get_vm_password($vm_info);
  $path = $vm_info->field_vm_path_name['und'][0]['value'];

  mercury_management_rediscover_vm_log($uow, "checking for FA agent installed\n");
  if ($uow['is_annotated'] == false) {
    $uow['cmds'][$idx + 1] = mercury_management_wcmd_vmrun_chk_for_fa($vc_hostname, $vc_username, $vc_password, $username, $password, $path);
  } else {
    $uow['results'][$idx + 1] = 'The file exists.';
    $uow['cmds'][$idx + 1] = "no_op";
  }

  $uow['cbks'][$idx + 1] = "mercury_management_wa_cb_disc_vm_vmrun_chk_for_fa";
  
  return true;
}


/**
 *  Next discovery step: vmrun output FA installed check now available.
 */
function mercury_management_wa_cb_disc_vm_vmrun_chk_for_fa(&$uow, $idx)
{
  $result = $uow['results'][$idx];
  $vm_info = $uow['vm_info'];
  $hostname = mercury_management_get_vm_hostname($vm_info);
  $username = mercury_management_get_vm_username($vm_info);
  $password = mercury_management_get_vm_password($vm_info);
  
  $driver_status = mercury_management_parse_vmrun_chk_for_fa($result);

  if (mercury_management_abort_updatevm_due_to_driver_status($uow, $driver_status)) {
    return false;
  }  
  if ($driver_status != FAMC_VM_DRIVER_STATUS_INSTALLED) {
    mercury_management_access_annotations('vm', $vm_info->title, 'clear');
    mercury_management_vm_discovery_successful($uow, $idx, 'VD');
    return true;
  }
  
  if ($uow['is_annotated'] == false) {
    mercury_management_access_annotations('vm', $vm_info->title, 'set');
  }

  mercury_management_rediscover_vm_log($uow, "obtaining cache options\n");

  $method = 'HgGetCacheOptions';
  $port   = mercury_management_get_vm_port($vm_info);
  $uow['cmds'][$idx + 1] = mercury_management_wcmd_mm_wsdl($hostname, $username, $password, $method, null, $port);
  $uow['cbks'][$idx + 1] = "mercury_management_wa_cb_disc_vm_getcacheoptions";

  return true;
}


/**
 *  Next discovery step: HgGetCacheOptions output now available.
 */
function mercury_management_wa_cb_disc_vm_getcacheoptions(&$uow, $idx)
{
  $result = implode("\n", $uow['results'][$idx]);
  eval("\$attrInfo = $result;");

  $vm_info = $uow['vm_info'];
  if ($attrInfo != 'error_wsdl_call') {
    $vm_info->field_fa_metrics_reset_interval['und'][0]['value'] = $attrInfo['metrics_reset_interval'];
    $uow['vm_info'] = $vm_info;
  }

  $hostname = mercury_management_get_vm_hostname($vm_info);
  $username = mercury_management_get_vm_username($vm_info);
  $password = mercury_management_get_vm_password($vm_info);

  mercury_management_rediscover_vm_log($uow, "obtaining cache info\n");

  $method = 'HgGetCacheInfo';
  $port   = mercury_management_get_vm_port($vm_info);
  $uow['cmds'][$idx + 1] = mercury_management_wcmd_mm_wsdl($hostname, $username, $password, $method, null, $port);
  $uow['cbks'][$idx + 1] = "mercury_management_wa_cb_disc_vm_getcacheinfo";

  return true;
}


/**
 *  Next discovery step: HgGetCacheInfo output now available.
 */
function mercury_management_wa_cb_disc_vm_getcacheinfo(&$uow, $idx)
{
  global $user;

  $result = implode("\n", $uow['results'][$idx]);
  eval("\$attrInfo = $result;");

  $vm_info = $uow['vm_info'];
  
  $vDiskArr = $uow['vDiskArr'];

  if ($attrInfo == 'error_wsdl_call') {
    if ($uow['is_annotated'] == true) {
      # Corner case where the annotation may be falsely present
      $uow['is_annotated'] = false;
      $host_and_vc_of_vm_nids =  mercury_management_get_host_dc_vc_of_vm($vm_info->nid);
      $vc_of_vm_node = node_load($host_and_vc_of_vm_nids['vcenter']);
      $vc_hostname = $vc_of_vm_node->field_hostname['und'][0]['value'];
      $vc_username = fmc_cipher::decrypt($vc_of_vm_node->field_username['und'][0]['value']);
      $vc_password = fmc_cipher::decrypt($vc_of_vm_node->field_password['und'][0]['value']);
      $username = mercury_management_get_vm_username($vm_info);
      $password = mercury_management_get_vm_password($vm_info);
      $path = $vm_info->field_vm_path_name['und'][0]['value'];

      mercury_management_rediscover_vm_log($uow, "checking for FA agent installed\n");

      $uow['cmds'][$idx + 1] = mercury_management_wcmd_vmrun_chk_for_fa($vc_hostname, $vc_username, $vc_password, $username, $password, $path);
      $uow['cbks'][$idx + 1] = "mercury_management_wa_cb_disc_vm_vmrun_chk_for_fa";
      return true;
    } else {
      $uow['error'] = "vm webservice call failed!";
      mercury_management_rediscover_vm_log($uow, $uow['error'] . "\n");
      mercury_utilities_fmc_log_message('VM Refresh', t('Error obtaining cache information from VM @vm.', array('@vm' => $vm_info->title)), WATCHDOG_ERROR);
      mercury_management_update_driver_status_on_phase_a_failure($uow, FAMC_VM_DRIVER_STATUS_NOT_KNOWN);
      return false;
    }
  } 

  $uow['attrInfo'] = $attrInfo;
  mercury_utilities_fmc_log_message('VM Refresh', t('Successfully obtained cache information from VM @vm.', array('@vm' => $vm_info->title)));
  $vm_info->field_fa_build_version['und'][0]['value'] = $attrInfo['build_version'];
  $vm_info->field_fa_cache_mode['und'][0]['value'] = $attrInfo['cache_mode'];
  $vm_info->field_fa_administration_state['und'][0]['value'] = $attrInfo['admin_state'];
  $vm_info->field_fa_operational_state['und'][0]['value'] = $attrInfo['op_state'];
  $vm_info->field_fa_operational_message['und'][0]['value'] = $attrInfo['op_message'];
  $vm_info->field_fa_performance_state['und'][0]['value'] = $attrInfo['perf_state'];
  $vm_info->field_fa_usable_blocks['und'][0]['value'] = $attrInfo['usable_blocks'];

  /*
   * Validate that the identifier the service thinks is the cache is the one added
   * to the VM as the cache device .. if not we need to notify the VM of the correct
   * cache space ID
   */
  $vm_cachespace_id = '';
  if (!empty($uow['cacheDevice'])) {
    $parts = explode('.', $uow['cacheDevice']);
    $vm_cachespace_id = $parts[1];
  }

  // Compare and make sure VM thinks the same
  $i1 = trim($attrInfo['cachespace_id']);
  $i2 = trim($vm_cachespace_id);
  if (!empty($i1) && !empty($i2) && ($i1 != $i2)) {
    mercury_management_rediscover_vm_log($uow, "synchronizing cache space ids $i1 vs $i2!\n");
    
    // If necessary, Disable cache before updating the ID
    if ($attrInfo['admin_state'] == "ADMIN_ENABLED")  {
      $param = array('request' => array('CacheState' => 'ADMIN_DISABLED'));
      $result = mercury_management_cache_operation('nid', $vm_info->nid, 'HgSetCacheState', $param);
    }
    $setcache_parameter = array('request' => array('CacheIdentifier' => $vm_cachespace_id));
    mercury_management_cache_operation('nid', $vm_info->nid, 'HgSetCacheIdentifier', $setcache_parameter);
    if ($attrInfo['admin_state'] == "ADMIN_ENABLED")  {
      $param = array('request' => array('CacheState' => 'ADMIN_ENABLED'));
      $result = mercury_management_cache_operation('nid', $vm_info->nid, 'HgSetCacheState', $param);
    }
    mercury_utilities_fmc_log_message('VM Refresh', t('Updated VM @vm cachespace ID to "@id"', array('@vm' => $vm_info->title, '@id' => $vm_cachespace_id)));
  }

  if (!empty($vm_cachespace_id)) {

    /**
     * Set the vmotion settings for the VM
     * If cache is allocated to the VM, find the hostlist. If there is more than one host
     * set the hostlist and vmotion = true. Also set the policy as custom.
     */
    $hosts = mercury_management_findCacheHosts($vm_info->nid, trim($vm_cachespace_id));
    if (count($hosts) > 0) {
      $vm_info->field_vm_vmotion_hostlist['und'][0]['value'] = serialize($hosts);
      if (count($hosts) > 1) {
	$vm_info->field_vm_vmotionable['und'][0]['value'] = 1;
      } else {
	$vm_info->field_vm_vmotionable['und'][0]['value'] = 0;
      }
      if (!empty($vm_info->field_vm_vmotion_policy['und'][0]['value'])) {
	$vm_info->field_vm_vmotion_policy['und'][0]['value'] = 'custom';
      }
    }
  }

  # Per the spec, we skip the vdisk info unless the VM has cache space 
  # allocated.  For now, since we have the vdisk info anyway, go ahead
  # and create the nodes.
  # if ($attrInfo['cache_size']) {
    mercury_management_vm_discovery_successful($uow, $idx, 'VD+CD+DD');
  # } else {
  #   mercury_management_vm_discovery_successful($uow, $idx, 'VD+CD');
  # }
  
  return true;
}


/**
 *  Perform Phase B (Database Update) for the given VM
 */
function mercury_management_discover_vm_db_commit(&$uow, $data_level)
{
   /*
    * Differences between VD and VD+CD are dealt with by 
    * mercury_management_add_host_control_1_data_to_host_node; all we have 
    * to do is create the VM node, and create/update or delete the vdisk
    * nodes.
    */

   mercury_management_rediscover_vm_log($uow, "committing vm to database ($data_level)\n");
   $vm_info = $uow['vm_info'];
   node_save($vm_info);
   if ($data_level == 'VD+CD+DD') {
      mercury_management_rediscover_vm_log($uow, "committing vdisk nodes to database\n");
      $attrInfo = $uow['attrInfo'];
      mercury_management_update_vm_vdisk_db_commit($uow, $attrInfo, $vm_info->nid);
   } else {
      if (false == $uow['vm_info_is_new']) {
         mercury_management_rediscover_vm_log($uow, "removing vdisk nodes from database\n");
         mercury_management_update_vm_vdisk_db_commit($uow, array(), $vm_info->nid);
      }
   }
}
